/*
 * $Header: /cvsroot/xwinx/xwinx/xwx8bit.c,v 1.7 2001/11/01 07:33:34 snaund Exp $
 * xwx8bit.c - Create 15bit->8bit color table
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <wingdi.h>
#include <getopt.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <pthread.h>
#include "xwx.h"
#include "xwxmemory.h"
#include "xwxmouse.h"
#include "xwxprint.h"
#include "xwxkeyboard.h"
#include "xwxhttp.h"
#include "xwxlines.h"
#include "xwx8bit.h"

static unsigned char closestindex(
		unsigned short r,
		unsigned short g, 
		unsigned short b,
		unsigned short *reds,
		unsigned short *greens,
		unsigned short *blues)
{
	unsigned char best;
	double bestvl;
	double vl;
	int i;

	best=0;
	bestvl = 
		((double)r - (double)reds[0]) *
		((double)r - (double)reds[0])
		+
		((double)g - (double)greens[0]) *
		((double)g - (double)greens[0])
		+
		((double)b - (double)blues[0]) *
		((double)b-(double)blues[0]);

	for (i = 1; i < 256; i++)
	{
		vl=
		((double)r - (double)reds[i]) *
		((double)r - (double)reds[i])
		+
		((double)g - (double)greens[i]) *
		((double)g - (double)greens[i])
		+
		((double)b - (double)blues[i]) *
		((double)b - (double)blues[i]);
		if (vl < bestvl)
		{
			best = i;
			bestvl = vl;
		}
	}

	return(best);
}

int mkcolors_8bit(struct x11info *x11info) 
{
	/* we're going to build a 15bit->8bit translation table,
	 * where the 15bits is a win32 0RRRRRGGGGGBBBBB and the 8bits
	 * is the colormap index
	 */
	unsigned char *temp;
	int i;
	XColor color;
	unsigned short r, g, b;
	unsigned short reds[256];
	unsigned short greens[256];
	unsigned short blues[256];

	temp = xwxcalloc(1, 32768);
	if (!temp)
		return(-1);

	freeup(x11info->colorindex);
	x11info->colorindex = temp;

	for (i = 0; i < 256; i++)
	{
		color.pixel = i;
		XQueryColor(x11info->display, x11info->colormap, &color);
		
		reds[i]		=	color.red;
		greens[i]	=	color.green;
		blues[i]	=	color.blue;
	}

	for (r = 0; r < 32; r++) 
	for (g = 0; g < 32; g++) 
	for (b = 0; b < 32; b++) {
		temp[ r * 1024 + g * 32 + b ] = closestindex(
			r << 11,
			g << 11,
			b << 11,
			reds,
			greens,
			blues
		);
	}

	return(0);
}
