/*
 * $Header: /cvsroot/xwinx/xwinx/xwxhttp.c,v 1.15 2001/11/01 07:33:34 snaund Exp $
 * xwxhttp.c - listen on a TCP port for HTTP requests for IPC
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <process.h>
#include <pthread.h>
#include <windows.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include "xwx.h"
#include "xwxhttp.h"
#include "xwxpng.h"
#include "xwxb64.h"

struct startstruct {
	char *httpauth;
	char *httpauth64;
};

static pthread_t pthread;

/* there are some possible problems with pthreads, so mutex's aren't
 * currently used here */
#if MUTEXWORKS
static pthread_mutex_t lock;
#endif
static int httpdport;
static int httpdfd;
static struct startstruct gl_ss;

int pause_xwinx(void);
int unpause_xwinx(void);

#define write_msg(a, b) write_msgn(a, b, strlen(b))

/* write *data to fd (upto len) */
static int write_msgn(int fd, char *data, int len)
{
	int n;

	/* while there is more data to write */
	while (len > 0)
	{
		/* write data to file descriptor (n = bytes written) */
		n = write(fd, data, len);

		/* there was a problem if *no* data was written */
		if (n < 1)
			return -1;

		/* make data start n bytes after it's previous location */
		data += n;

		/* length is now shorter by the n bytes that were written */
		len -= n;
	}
	return 0;
}

/* read a single character from a socket -- if an error happens,
 * just return '\n' */
static char read_char(int fd, time_t stop) 
{
	fd_set rset;
	time_t t1;
	struct timeval tv;
	char data[1];
	int t;
	int n;
	int i;

	t1 = time(NULL);
	t = stop - t1;
	if (t < 0)
		goto error;
	t++;

	FD_ZERO(&rset);
	FD_SET(fd, &rset);
	tv.tv_sec = t;
	tv.tv_usec = 0;

	i = select(fd + 1, &rset, NULL, NULL, &tv);
	if (i != 1)
		goto error;

	n = read(fd, data, 1);
	if (n < 1)
		goto error;

	return data[0];

	error:
		return '\n';
}

/* this sends a pause or unpause instruction to another copy of xwinx
 * via http */
static int dopause_httpd(int port, int mode) 
{
	int fd;
	int i;
	struct sockaddr_in sa;

	memset(&sa, 0, sizeof(struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = 16777343; /* 127.0.0.1 in little endian */

	fd=socket(AF_INET, SOCK_STREAM, 0);
	if (fd == -1)
		goto error;
	i = connect(fd, (struct sockaddr*)&sa, sizeof(struct sockaddr_in));
	if (i)
		goto error;

	if (!mode)
		write_msg(fd,"GET /pause\r\n\r\n");
	else
		write_msg(fd,"GET /unpause\r\n\r\n");

	read_char(fd, time(NULL) + 2);

	close(fd);
	return 0;
	error:
		if (fd!=-1) close(fd);
		fprintf(stderr, "Error sending pause/unpause request\n");
		return -1;
}

/* just here for readability */
int pause_httpd(int port)
{
	return dopause_httpd(port,0);
}

int unpause_httpd(int port)
{
	return dopause_httpd(port,1);
}

/* generic httpd responses */
static void reply_httpd(int fdin, int valid)
{
	if (valid)
	{
		write_msg(
			fdin,
			"HTTP/1.0 200 OK\r\n"\
			"Content-type: xwinx/status\r\n"\
			"Content-length: 7\r\n"\
			"\r\n"\
			"SUCCESS"
		);
	} else
	{
		write_msg(
			fdin,
			"HTTP/1.0 200 OK\r\n"\
			"Content-type: xwinx/status\r\n"\
			"Content-length: 7\r\n"\
			"\r\n"\
			"FAILURE"
		);
	}
}

/* request a login:password */
static void reply401_httpd(int fdin)
{
	write_msg(
		fdin,
		"HTTP/1.0 401 Authorization Required\r\n"\
		"WWW-Authenticate: Basic realm=\"XWinX\"\r\n"\
		"Content-Type: text/html\r\n"\
		"\r\n"\
		"<HTML><HEAD>\n"\
		"<TITLE>401 Authorization Required</title>\n"\
		"</head><BODY>\n"\
		"<H1>401 Authorization Required</h1>\n"\
		"</body></html>\n"
	);
}

/* send a current screenshot of the display as a .png file */
static void replypng_httpd(int fdin)
{
	struct pngdata pngdata;

	memset(&pngdata, 0, sizeof(struct pngdata));

	/* get a screencap as a png */
	getscreen_png(&pngdata);

	/* send the png to the web browser */
	if (pngdata.error)
	{
		reply_httpd(fdin,0);
	} else
	{
		char buff[15];

		snprintf(buff, 15, "%d", pngdata.bytes);

		write_msg(
			fdin,
			"HTTP/1.0 200 OK\r\n"\
			"Content-type: image/png\r\n"\
			"Content-length: "
		);
		write_msg(fdin, buff);
		write_msg(
			fdin,
			"\r\n"\
			"\r\n"
		);
		write_msgn(
			fdin,
			pngdata.data,
			pngdata.bytes
		);
	}

	/* free up the PNG memory */
	if (pngdata.data)
		free(pngdata.data);
}

/* returns 1 if authentication is okay 
 * 
 * note that auth64=="" => no auth necessary, but auth64==NULL => deny service
 */
static int checkauth(char *basic, char *auth64)
{

	if (!auth64)
		return 0;
	if (!auth64[0])
		return 1;
	if (!strcmp(basic, auth64))
		return 1;
	return 0;
}

/* handles and processes and HTTP request */
static void handlerequest(int fdin, int local, char *auth64)
{
	char	oneline[80];
	char	uri[80] = "";
	char	basic[80] = "";
	int	i;
	int	bytes;
	time_t	stop;
	int	valid=1;
	char	*temp;
	int	auth=0;

	i = 0;
	bytes = 0;
	stop = time(NULL) + 5;

	while (1)
	{
		if (i == 79)
			i = 0;
		
		if (bytes == 4 * 1024)
			break;
		
		oneline[i] = read_char(fdin, stop);
		bytes++;
		
		if (oneline[i] == '\r')
			continue;

		if (oneline[i]!='\n')
		{
			i++;
		} else
		{
			if ( (i) && (oneline[i-1] == '\r') )
				oneline[i-1] = '\0';
			else
				oneline[i] = '\0';

			if (!oneline[0])
				break;

			if (!strncasecmp(oneline, "GET ", 4))
			{
				temp = strchr(oneline + 4, ' ');
				if (temp)
					temp[0] = '\0';
				strncpy(uri, oneline + 4, 79);
				uri[79] = '\0';
			}

			if (!strncasecmp(oneline, "Authorization: Basic ", 21))
			{
				strncpy(basic, oneline + 21, 79);
				basic[79] = '\0';
			}

			i = 0;
		}
	}

	auth = checkauth(basic, auth64);

	if (!local && !auth)
	{
		reply401_httpd(fdin);
		return;
	}

	if (!strcmp(uri,"/pause"))
	{
		pause_xwinx();
	} else if (!strcmp(uri,"/unpause"))
	{
		unpause_xwinx();
	} else if (!strcmp(uri,"/screen.png"))
	{
		/* return after the reply */
		if (auth)
			replypng_httpd(fdin);
		else
			reply401_httpd(fdin);
		return;
	} else
		valid=0;

	reply_httpd(fdin, valid);
}

static void *listen_httpd(void *ignore)
{
	int	port;
	int	fd;
	struct	sockaddr_in sa;
	struct	startstruct startstruct;
	int	ssa;
	int	fdin;
	int	local;

	port = httpdport;
	fd = httpdfd;

	/* this one prints my hostname string...
	fprintf(
		stderr,
		"In: %s\n",
		((struct startstruct *)startup_in)->httpauth
	);
	fprintf(
		stderr,
		"In: %s\n",
		((struct startstruct *)startup_in)->httpauth64
	);
	*/
	memcpy(&startstruct, &gl_ss, sizeof(struct startstruct));

#if MUTEXWORKS
	/* un-deadlock the creator */
	pthread_mutex_unlock(&lock);
#endif

	while (1)
	{
		ssa = sizeof(struct sockaddr_in);
		fdin = accept(fd, (struct sockaddr*)&sa, &ssa);
		if (fdin == -1)
		{
			if (errno == EBADF)
				break;
			sleep(1);
			continue;
		}

		local = 0;

		if (sa.sin_addr.s_addr == 16777343) {
			local = 1;
		}
		handlerequest(fdin, local, startstruct.httpauth64);
		close(fdin);
		sleep(1);
	}

	freeup(startstruct.httpauth);
	freeup(startstruct.httpauth64);
	return 0;
}

void deinit_httpd(void)
{
	/* TODO:  kill the thread easily 
	 * note that it is at least mildly tricky
	 * */
}

static char *getauth64(char *httpauth)
{
	char	*temp=NULL;
	int	n;

	if (!httpauth)
		return NULL;

	n = strlen(httpauth);

	temp = malloc(n * 2 + 4);
	if (!temp)
		return NULL;

	encode_base64(temp,httpauth,n);

	return temp;
}

int init_httpd(struct params* params, int port)
{
	struct	sockaddr_in sa;
	struct	startstruct startstruct;
	int	fd;
#if MUTEXWORKS
	int	inim=0;
#endif

	if (!port)
		return 0;

	/* initialise the startup structure for listen_httpd() */ 
	memset(&startstruct, 0, sizeof(struct startstruct));

	/* initialise the sa structure */
	memset(&sa, 0, sizeof(struct sockaddr_in));

	/* open fd as a TCP/IP socket */
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd==-1)
		goto error;

	/* setup the sockaddr_in scructure */
	sa.sin_port = htons(port);
	sa.sin_family = AF_INET;

	/* bind to a port */
	if (bind(fd, (struct sockaddr*)&sa, sizeof(struct sockaddr_in)))
	{
		fprintf(stderr, "Error binding to port %d\n", port);
		goto error;
	}

	if (listen(fd,	5))
		goto error;

	httpdfd=fd;
	httpdport=port;

	/* if a user/pass was specified, find the base64 representation and set
	 * the members in startstruct
	 */
	if (params->httpauth)
	{
		startstruct.httpauth = strdup(params->httpauth);
		if (!startstruct.httpauth)
			goto error;
		startstruct.httpauth64 = getauth64(startstruct.httpauth);
		if (!startstruct.httpauth64)
			goto error;
	}

	/* copy to global space for debugging */
	memcpy(&gl_ss, &startstruct, sizeof(struct startstruct));
	
#if MUTEXWORKS
	pthread_mutex_init(&lock, NULL);
	/* we use the fact that a double lock is a deadlock
	 * for pthreads to ensure that the created thread can get
	 * the data out before we continue
	 *
	 * perhaps not all pthreads can do this
	 */
	pthread_mutex_lock(&lock);
#endif

	/* create a thread to listen for HTTP, pass a pointer to startstruct */
	if (pthread_create(&pthread, NULL, listen_httpd, (void *)&startstruct))
	{
		fprintf(stderr, "Error creating httpd thread\n");
		goto error;
	}

#if MUTEXWORKS
	pthread_mutex_lock(&lock);

	inim = 0;
	pthread_mutex_unlock(&lock);
	pthread_mutex_destroy(&lock);
#endif

	printf("httpd thread created on port %d\n", port);

	return 0;
error:
	freeup(startstruct.httpauth);
	freeup(startstruct.httpauth64);
	if (fd != -1) close(fd);
#if MUTEXWORKS
	if (inim)
	{
		pthread_mutex_unlock(&lock);
		pthread_mutex_destroy(&lock);
	}
#endif
	return -1;
}
