/*
 * $Header: /cvsroot/xwinx/xwinx/xwxkeyboard.c,v 1.14 2001/11/08 00:10:36 guilty Exp $
 * xwxkeyboard.c - xwinx keyboard support functions 
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <wingdi.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <pthread.h>
#include "xwx.h"
#include "xwxprint.h"
#include "xwxwindows.h"
#include "xwxkeyboard.h"

/* convert a position in the modifer keycode array into the corresponding
 * mask for that modifier
 */
int keyModposToModmask (int pos, int pospermod)
{
	switch (pos / pospermod)
	{
		case 0:
			return(ShiftMask);
		case 1:
			return(LockMask);
		case 2:
			return(ControlMask);
		case 3:
			return(Mod1Mask);
		case 4:
			return(Mod2Mask);
		case 5:
			return(Mod3Mask);
		case 6:
			return(Mod4Mask);
		case 7:
			return(Mod5Mask);
	}
	return(0);
}

/* this function translates a-zA-Z0-9 */
int keyOffset (KeySym xks)
{
	/* where the key is inside a range that is sequential
	 * in both the win32 and X11 keycodes, just change the 
	 * offsett to be relative to the windows offset
	 */
	if (xks >= XK_A	&& xks <= XK_Z)
		return(xks - XK_A + XWX_VK_A);

	if (xks >= XK_a	&& xks <= XK_z)
		return(xks - XK_a + XWX_VK_A);

	if (xks >= XK_0	&& xks <= XK_9)
		return(xks - XK_0 + XWX_VK_0);

	if (xks >= XK_KP_0 && xks <= XK_KP_9)
		return(xks - XK_KP_0 + VK_NUMPAD0);

	if (xks >= XK_F1 && xks <= XK_F24)
		return(xks - XK_F1 + VK_F1);
	
	return(0);
}

/* convert between an xlib keycode and a win32 virtual keycode */
int keyTranslate (KeySym xks)
{
	switch (xks)
	{
		case XK_Alt_L:
		case XK_Alt_R:
			return(VK_MENU);

		case XK_Control_R:
		case XK_Control_L:
			return(VK_CONTROL);

		case XK_Shift_L:
		case XK_Shift_R:
			return(VK_SHIFT);

		case XK_Caps_Lock:
			return(VK_CAPITAL);

		case XK_Scroll_Lock:
			return(VK_SCROLL);

		case XK_Num_Lock:
			return(VK_NUMLOCK);

		case XK_BackSpace:
			return(VK_BACK);

		case XK_Tab:
			return(VK_TAB);

		case XK_KP_Enter:
		case XK_Return:
			return(VK_RETURN);

		case XK_Escape:
			return(VK_ESCAPE);

		case XK_Insert:
			return(VK_INSERT);

		case XK_Delete:
			return(VK_DELETE);

		case XK_Left:
			return(VK_LEFT);

		case XK_Right:
			return(VK_RIGHT);

		case XK_Up:
			return(VK_UP);

		case XK_Down:
			return(VK_DOWN);

		case XK_Page_Up:
			return(VK_PRIOR);

		case XK_Page_Down:
			return(VK_NEXT);

		case XK_Home:
			return(VK_HOME);

		case XK_End:
			return(VK_END);

		case XK_Print:
			return(VK_SNAPSHOT);

		case XK_Pause:
			return(VK_PAUSE);

		case XK_space:
			return(VK_SPACE);

		case XK_Meta_L:
			return(VK_LWIN);

		case XK_Menu:
			return(VK_APPS);
	}

	return(0);
}

/* converts a KeySym to a char of what's on the key for VkKeyScan to use */
char keyCharacter (KeySym xks)
{
	switch (xks)
	{
		case XK_plus:
			return('+');

		case XK_comma:
			return(',');

		case XK_minus:
			return('-');

		case XK_period:
			return('.');

		case XK_semicolon:
			return(';');

		case XK_apostrophe:
			return('\'');

		case XK_slash:
			return('/');

		case XK_grave:
			return('`');

		case XK_bracketleft:
			return('[');

		case XK_bracketright:
			return(']');

		case XK_backslash:
			return('\\');

		case XK_equal:
			return('=');
	}
	
	return(0);
}

/* translate modifier keys (CAPS and CTRL) */
int keyModifier (unsigned int code, XModifierKeymap *map)
{
	int i;

	/* loop through the modifier map array */
	for (i = 0; i < map->max_keypermod * 8; i++)
	{
		if (code != map->modifiermap[i])
			continue;

		/* check what modifier set the kecode belongs to */
		switch (i / map->max_keypermod)
		{
			case 1:
				return(VK_CAPITAL);
			case 2:
				return(VK_CONTROL);
		}
	}
	return(0);
}

/* translate keys on the numeric keypad */
int keyNumeric (KeySym xks, int numlock)
{
	switch (xks)
	{
		case XK_KP_Divide:
			return(VK_DIVIDE);

		case XK_KP_Multiply:
			return(VK_MULTIPLY);

		case XK_KP_Subtract:
			return(VK_SUBTRACT);

		case XK_KP_Add:
			return(VK_ADD);

		case XK_KP_Delete:
			return(numlock ? VK_DECIMAL : VK_DELETE);

		case XK_KP_Insert:
			return(numlock ? VK_NUMPAD0 : VK_INSERT);

		case XK_KP_End:
			return(numlock ? VK_NUMPAD1 : VK_END);

		case XK_KP_Down:
			return(numlock ? VK_NUMPAD2 : VK_DOWN);

		case XK_KP_Next:
			return(numlock ? VK_NUMPAD3 : VK_NEXT);

		case XK_KP_Left:
			return(numlock ? VK_NUMPAD4 : VK_LEFT);

		case XK_KP_Begin:
			return(numlock ? VK_NUMPAD5 : 0);

		case XK_KP_Right:
			return(numlock ? VK_NUMPAD6 : VK_RIGHT);

		case XK_KP_Home:
			return(numlock ? VK_NUMPAD7 : VK_HOME);

		case XK_KP_Up:
			return(numlock ? VK_NUMPAD8 : VK_UP);

		case XK_KP_Prior:
			return(numlock ? VK_NUMPAD9 : VK_PRIOR);
	}
	return(0);
}

/* function to press a key in windows, virtual keycode is first argument */
void keyWin32Press (int keycode)
{
	/* a keyboard down event will unpause xwinx if it is paused */
	unpause_xwinx();
	
	keybd_event(
		keycode,
		MapVirtualKey(keycode, 0),
		(DWORD)NULL,
		0
	);
}

/* release a key in windows */
void keyWin32Release (int keycode)
{
	keybd_event(
		keycode,
		MapVirtualKey(keycode, 0),
		KEYEVENTF_KEYUP,
		0
	);
}

/* wrapper to convert a key event into a windows virtual keycode */
int keyConvert (struct xinput *xinput, struct x11info *x11info, 
	struct params *params, XEvent evt)
{
	KeySym	sym;		/* Key symbol value for event */
	KeyCode num;		/* num lock key code */
	int	msk = 0;	/* num lock modifier mask */
	int	wky;		/* windows virtual key code */
	int	wch;		/* character for keyboard key */
	int	nlk = 0;	/* numlock on or off */
	int	i;

	/* get the KeySym (key symbol) for the key event */
	pthread_mutex_lock(&xinput->inputlock);
	sym = XKeycodeToKeysym(x11info->display, evt.xkey.keycode, 0);
	pthread_mutex_unlock(&xinput->inputlock);

	/* converts modifier keys (control and caps lock) */ 
	wky = keyModifier(evt.xkey.keycode, x11info->modifiers);
	if (wky)
		return(wky);

	/* this translates aplanumeric keys */
	wky = keyOffset(sym);
	if (wky)
		return(wky);
	
	/* this translates non-printable keys mainly */
	wky = keyTranslate(sym);
	if (wky)
		return(wky);

	/* this converts printable, non-letter characters (punctuation) */
	wch = keyCharacter(sym);
	if (wch)
		return(VkKeyScan(wch));

	/* find the modifermask for num lock */
	pthread_mutex_lock(&xinput->inputlock);
	num = XKeysymToKeycode(x11info->display, XK_Num_Lock);
	pthread_mutex_unlock(&xinput->inputlock);

	for (i = 0; i < x11info->modifiers->max_keypermod * 8; i++)
	{
		if (num == x11info->modifiers->modifiermap[i])
			msk = keyModposToModmask(
				i,
				x11info->modifiers->max_keypermod
			);

		if (msk)
			break;
	}	

	nlk = ( (evt.xkey.state & msk) == msk ) ? 1 : 0;

	/* numeric keypad keys */
	wky = keyNumeric(sym, nlk);
	if (wky)
		return(wky);

	/* warn that the translation failed */
	pthread_mutex_lock(&xinput->inputlock);
	wrnKeyboard(params, XKeysymToString(sym));
	pthread_mutex_unlock(&xinput->inputlock);

	return(0);
}

/* simulate a corresponding win32 keypress for an XEvent if appropriate */
void keyPressEvent (struct xinput *xinput, struct x11info *x11info, 
	struct params *params, XEvent evt)
{
	int vkc;		/* virtual keycode */

	/* return unless this is a KeyPress or KeyRelease function */
	if (evt.type != KeyPress && evt.type != KeyRelease)
		return;

	/* find the windows virtual keycode for this event */
	vkc = keyConvert(xinput,x11info, params, evt);

	/* simulate the event if the keycode was translated OK */
	if (vkc && evt.type == KeyPress)
		keyWin32Press(vkc);

	if (vkc && evt.type == KeyRelease)
		keyWin32Release(vkc);
}


/* simulate a corresponding win32 keypress for an XEvent if appropriate */
void keySyncLocks (struct xinput *xinput, struct x11info *x11info, 
	struct params *params, XEvent evt)
{
	int i;	/* counter for the for loop */

	/* exit the function if this is not a key event */
	if (evt.type != KeyPress)
		return;

	/* loop through the 3 locks we are checking (caps, scroll, num) */
	for (i = 0; i < 8 * x11info->modifiers->max_keypermod; i++)
	{
		KeySym sym;
		int msk = 0;
		int vkc = 0;

		pthread_mutex_lock(&xinput->inputlock);
		sym = XKeycodeToKeysym(
			x11info->display,
			x11info->modifiers->modifiermap[i],
			0
		);
		pthread_mutex_unlock(&xinput->inputlock);

		switch (sym)
		{
			case XK_Scroll_Lock:
				vkc = VK_SCROLL;
				break;
			case XK_Num_Lock:
				vkc = VK_NUMLOCK;
				break;
		}

		msk = keyModposToModmask(i, x11info->modifiers->max_keypermod);

		if (msk == LockMask)
			vkc = VK_CAPITAL;
	
		if (!vkc)
			continue;

		if (
	 		(GetKeyState(vkc) & 1)
			!=
		 	((evt.xkey.state & msk) == msk)
		)
		{
			keyWin32Press(vkc);
			keyWin32Release(vkc);
		}
	}
}
