/*
 * $Header: /cvsroot/xwinx/xwinx/xwxpng.c,v 1.13 2001/11/08 21:59:47 guilty Exp $
 * xwxpng.c - save the screen as a png
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * most of the includes are just for definitions for xwx.h
 * TODO: split xwx.h
 */

#define PNG_STATIC 1
#define ZLIB_STATIC 1

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <process.h>
#include <pthread.h>
#include <windows.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <zlib.h>
#include <png.h>
#include "xwx.h"
#include "xwxhttp.h"
#include "xwxmemory.h"
#include "xwxpng.h"


/* callback function for libpng -- "writes" the data by
 * adding it to a buffer in ram
 */
void user_write_fn(png_structp png_ptr, png_bytep data, png_size_t length)
{
	struct pngdata *ptr;
	char *temp;

	ptr = png_get_io_ptr(png_ptr);

	if (ptr->error)
		return;

	temp = realloc(ptr->data, ptr->bytes + length);
	if (!temp)
		ptr->error=-1;

	ptr->data = temp;

	memcpy(ptr->data+ptr->bytes,data,length);

	ptr->bytes += length;
}

/* since we spool the data into ram, there's no need for a flush
 */
void user_IO_flush_function(png_structp png_ptr)
{
}

/* libpng will call this if it encounters an error from which it
 * can't continue
 */
void error_fn(png_structp png_ptr, png_const_charp error_msg)
{
	fprintf(stderr,"libpng had an error (exiting)\n");

	/* WARNING we actually exit on this error as libpng does not
	 * support continuing after this point (???)
	 *
	 * we should use setjmp() or similar to combat this
	 * What works in win32 for this?
	 * */
	exit(EXIT_FAILURE);
}

/* libpng will call this if it encounters an error.
 * this is ignored for now
 */
void warning_fn(png_structp png_ptr, png_const_charp warning_msg)
{
	struct pngdata *ptr;

	ptr = png_get_error_ptr(png_ptr);
	ptr->warn += 1;
}


/* we have the screen in pixels[], and we just need to pass it to libpng
 */
static int write_png(unsigned char *pixels, struct pngdata *wi, int w, int h)
{
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	int i, j, k;
	png_bytep row = NULL;
	unsigned char *rowt, *pixt;
	unsigned short *colors;
	int Bpl;

	Bpl = w * 2;

	memset(wi, 0, sizeof(struct pngdata));

	/* failure is acceptable, so don't do a xwxmalloc() which => exit() */
	row = malloc(3 * w);
	if (!row)
		goto error;

	/* some png init routines */

	png_ptr=png_create_write_struct(
		PNG_LIBPNG_VER_STRING,
		wi,
		error_fn,
		warning_fn
	);

	if (!png_ptr)
		goto error;

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
		goto error;

	png_set_write_fn(
		png_ptr,
		wi,
		user_write_fn,
		user_IO_flush_function
	);

	/* 8 bit depth =>24 bit colors here 
	 * compression values come from zlib
	 * Default is Z_DEFAULT_COMPRESSION
	png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	png_set_compression_level(png_ptr, Z_BEST_SPEED);
	png_set_compression_level(png_ptr, Z_NO_COMPRESSION);
	 * */

	png_set_IHDR(
		png_ptr,
		info_ptr,
		w,
		h,
		8,
		PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_BASE,
		PNG_FILTER_TYPE_BASE
	);

	png_write_info(png_ptr, info_ptr);

	/* we write the png one row at a time */

	pixt = pixels;
	for (i = 0; i < h; i++)
	{
		rowt = row;
		colors = (unsigned short *)pixt;
		for (j = 0; j < w; j++)
		{
			k = colors[j];
			rowt[0] = (k >> 7) & (248);
			rowt[1] = (k >> 2) & (248);
			rowt[2] = (k << 3) & (248);
		
			rowt += 3;
		}
		png_write_row(png_ptr, row);
		pixt += Bpl;
	}

	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	free(row);

	return 0;
	
error:
	png_destroy_write_struct(&png_ptr, &info_ptr);
	if (row)
		free(row);
	return -1;
}

int getscreen_png(struct pngdata *pngdata)
{
	int toret;
	HDC dev = NULL;
	HDC mdev=NULL;
	int w, h;
	BITMAPINFO bmi;
	HBITMAP bmap=NULL;
	HBITMAP omap;
	unsigned char *pixels = NULL;

	dev = GetDC(NULL);
	if (!dev)
		goto error;

	w = GetDeviceCaps(dev, HORZRES);
	h = GetDeviceCaps(dev, VERTRES);

	mdev=CreateCompatibleDC(dev);
	if (!mdev) goto error;
	
	bmap=CreateCompatibleBitmap(dev,w,h);
	if (!bmap) goto error;

	/* failure is acceptable, so don't do a xwxmalloc() which => exit() */
	/* if w is odd, this won't display properly, won't crash either */
	if ( (w & 1) )
		pixels = calloc( 1, (w + 1) * h * 2 );
	else
		pixels = calloc( 1, w * h * 2);
	if (!pixels)
		goto error;

	bmi.bmiHeader.biSize		= sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth		= w;
	bmi.bmiHeader.biHeight		= -h;
	bmi.bmiHeader.biPlanes		= 1;
	bmi.bmiHeader.biBitCount	= 16;
	bmi.bmiHeader.biCompression	= BI_RGB;
	bmi.bmiHeader.biSizeImage	= 0;
	bmi.bmiHeader.biXPelsPerMeter	= 0;
	bmi.bmiHeader.biYPelsPerMeter	= 0;
	bmi.bmiHeader.biClrUsed		= 0;
	bmi.bmiHeader.biClrImportant	= 0;

	/* this gets the image from windows */

	omap=SelectObject(mdev,bmap);
	if (!BitBlt(mdev, 0, 0, w, h, dev, 0, 0, SRCCOPY)) {
		goto error;
	}

	if (!GetDIBits(
		mdev,
		bmap,
		0,
		h,
		pixels,
		&bmi,
		DIB_RGB_COLORS
	)) {
		DWORD err;
		
		err=GetLastError();
		fprintf(stderr,"Error capturing screen: %lu\n",err);

	}

	SelectObject(mdev,omap);

	ReleaseDC(NULL, dev);
	ReleaseDC(NULL, mdev);
	DeleteObject(bmap);

	/* spool the png into a memory buffer in pngdata */
	toret = write_png(pixels, pngdata, w, h);
	free(pixels);

	/* pngdata will still have values for the caller to see
	 * (errors, data, etc.)
	 */

	return toret;
error:
	fprintf(stderr,"Error getting png image\n");
	if (pixels)
		free(pixels);
	if (dev)
		ReleaseDC(NULL, dev);
	if (mdev)
		ReleaseDC(NULL, mdev);
	if (bmap)
		DeleteObject(bmap);
	return -1;
}
